// Frequency counter (0.5 Hz to 40 MHz)
// Counts frequency on P0, sends ASCII to PC at 115200 Baud
// 1.0 Second gate time (gate time will be longer if (frequency is < 2 Hz)
//
#include <stdio.h>
#include <propeller.h>

// Disable the standard serial driver
_Driver *_driverlist[] = { NULL };

void SEROUT(int txpin, int baudrate, char *str);

#define BAUD   115200    // Baud rate for PC communications
#define Signal 15        // Input pin for signal
#define Mask   (1 << Signal)
#define TX     30        // Output pin for PC communications
#define FREQ   1234567   // Generated frequency
#define INCR   (FREQ*(1LL << 31)/40000000)

int cntStart;            // "cnt" value at start of measurement
int cntTime;             // elapsed "cnt" clocks
int sigCnt;              // count of how many input pulses have been received
int temp;                // counter
int digit;               // current digit
int digitSum;            // Used for leading zero removal
char ascii[20];          // Holds ascii representation of frequency

void main() {
  // Set up counter B to generate FREQ frequency
  CTRB = (0x4 << 26) | Signal;
  FRQB = INCR;
  DIRA |= Mask;
  
  // Set up counter A to count rising edges
  CTRA = (0xA << 26) | Signal;
  FRQA = 1;
  
  // Set serial transmit bit for output
  OUTA |= (1 << TX);  
  DIRA |= (1 << TX);  
  
  while (1) {
    // Wait for a signal pulse to syncronize to the system counter
    waitpne(Mask, Mask);
    waitpeq(Mask, Mask);
    PHSA = 0;
    cntStart = CNT - 4;

    // Count input pulses until 1.0 seconds have elapsed
    do {
      waitpne(Mask, Mask);
      waitpeq(Mask, Mask);
      sigCnt = PHSA;
      cntTime = CNT - cntStart;
    } while (cntTime <= 80000000); // 1.0 Second gate time
    cntTime = cntTime - 4; // To account for "sigCnt = phsa" instruction

    // Calculate  frequency if >= 0.5 Hz
    if (cntTime <= 160000000) {

      // Since cntTime is in 80MHz units, the frequency is sigCnt / (cntTime / 80_000_000) = Hertz
      // Rearranged this can be written as (sigCnt * 80_000_000) / cntTime = Hertz
      //
      // We want the result in milliHertz (1/1000 of a hertz) so we need to use:
      // (sigCnt * 80_000_000_000) / cntTime = milliHertz
      //
      // Okay, now we have a problem, 80_000_000_000 cannot even be represented in 32 bits. Hmmm
      // What if (we calculate the answer one digit at a time...
      //
      // The leftmost digit is MegaHertz && that digit can be calculated as:
      // (sigCnt * 80) / cntTime = MegaHertz. This we can do easily.
      //
      // Now the "trick" is how do we get the next digit ?
      // We take the remainder from the division, multiply it by 10 && divide again.
      // by repeating this process of taking the remainder && multipling by 10 we
      // can get as many digits as we need until we run out of precision.
      //
      digitSum = 0;                // Sum of all digits so far. Used to create replace leading zeros with spaces
      ascii[0] = 0;
      sigCnt = sigCnt * 8;         // Scale signal count to get MegaHertz digit

      for (temp = 0; temp <= 13; temp++) { // 7 digits + decimal point + 3 digits = 11 digits

        digit = sigCnt / cntTime;  // Calculate this digit of result
        sigCnt %= cntTime;         // Keep remainder for next digit

        sigCnt = (sigCnt << 1)     // sigCnt = sigCnt * 10
               + (sigCnt << 3);    // Scale signal count for next digit

        digitSum += digit;         // Add this digit's value to sum
        digit = digit + '0';       // Convert digit to an ascii digit

        // if (this is a leading zero, make it a space
        if (temp < 9 &&            // Only use leading spaces for digits higher than the units digit
         digitSum == 0) {          // if (the sum is zero, { we have only had "0" digits so far
          digit = ' ';             //  so use a space instead of a "0" for this digit.
        }

        ascii[temp] = digit;       // Put digit into the result string

        // Insert a comma if (needed
        if (temp == 1 ||
         temp == 5) {
          temp++;
          if (digitSum == 0)
            ascii[temp] = ' ';
          else
            ascii[temp] = ',';
        }

        // Insert decimal point at proper place
        if (temp == 9) {           // if (we just processed the units digit,
          temp++;                  //  move pointer to the next character in the result string
          ascii[temp] = '.';       //  and make it a decimal point
        }

      }                            // Process all digits

      strcpy(ascii+14, " Hz\r");

    } else {
      strcpy(ascii, "Too low (< 0.5Hz)\r");
    }

    SEROUT(TX, BAUD, ascii);      // Send frequency to terminal

  }                               // Repeat forever
}

void SEROUT(int txpin, int baudrate, char *str)
{
  int value;
  int txmask = (1 << txpin);
  int bitcycles = 80000000 / baudrate;
  int count = CNT;

  while ((value = *str++))
  {
    value = (value << 1) | 0x200;
    count = CNT;
    while (value)
    {
      if (value & 1)
        OUTA |= txmask;
      else
        OUTA &= ~txmask;
      waitcnt(count += bitcycles);
      value >>= 1;
    }      
  }
}

